@@ -11,7 +11,7 @@ Released under the MIT License.
 
 @author      Erki Suurjaak
 @created     05.03.2014
-@modified    21.11.2022
+@modified    24.11.2022
 ------------------------------------------------------------------------------
 """
 import collections
@@ -84,9 +84,9 @@ def select(table, cols="*", where=(), group=(), order=(), limit=(), **kwargs):
     @param   where   columns or expressions to select by, as dict or a sequence
                      of key-value tuples; value can specify operator
                      e.g. `{"mycolumn": ("IN", [1, 2, 3])}`
-    @param   cols    columns to select if not all, as string or a sequence of strings
-    @param   group   columns to GROUP BY, as string or a sequence of strings
-    @param   order   columns to ORDER BY, as string, a sequence of strings,
+    @param   cols    columns to select if not all, as string or a sequence of stringables
+    @param   group   columns to GROUP BY, as string or a sequence of stringables
+    @param   order   columns to ORDER BY, as string, a sequence of stringables,
                      or a combination of column names and direction flags
     @param   limit   LIMIT .. OFFSET.. values, as integer or a sequence of integers;
                      None or -1 disables LIMIT or OFFSET
@@ -112,7 +112,12 @@ def delete(table, where=(), **kwargs):
 
 
 def execute(sql, args=()):
-    """Executes the SQL statement and returns database cursor."""
+    """
+    Executes the SQL statement and returns database cursor.
+
+    @param   sql   SQL statement to execute, with engine-specific parameter bindings, if any
+    @param   args  query parameters, as tuple or dictionary
+    """
     return init().execute(sql, args)
 
 
@@ -122,6 +127,8 @@ def executescript(sql):
 
     Note that in SQLite, the statements are executed outside of transaction,
     and any pending transaction will be committed first.
+
+    @param   sql   script with one or more SQL statements
     """
     return init().executescript(sql)
 
@@ -245,7 +252,12 @@ class Queryable(object):
 
 
     def execute(self, sql, args=()):
-        """Executes the SQL statement and returns database cursor."""
+        """
+        Executes the SQL statement and returns database cursor.
+
+        @param   sql   SQL statement to execute, with engine-specific parameter bindings, if any
+        @param   args  query parameters, as tuple or dictionary
+        """
         raise NotImplementedError()
 
 
@@ -255,12 +267,13 @@ class Queryable(object):
 
         Note that in SQLite, the statements are executed outside of transaction,
         and any pending transaction will be committed first.
+
+        @param   sql   script with one or more SQL statements
         """
         raise NotImplementedError()
 
 
-    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(),
-                limit=(), values=()):
+    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(), limit=(), values=()):
         """Returns (SQL statement string, parameter dict)."""
         raise NotImplementedError()
 
@@ -271,12 +284,19 @@ class Queryable(object):
         raise NotImplementedError()
 
 
+    @property
+    def cursor(self):
+        """Database engine cursor object, or `None` if closed."""
+        raise NotImplementedError()
+
+
     @classmethod
     def quote(cls, value, force=False):
         """
         Returns identifier in quotes and proper-escaped for queries,
         if value needs quoting (has non-alphanumerics, starts with number, or is reserved).
 
+        @param   value  the value to quote, returned as-is if not string
         @param   force  whether to quote value even if not required
         """
         raise NotImplementedError()
@@ -350,6 +370,7 @@ class Transaction(Queryable):
 
         Note that parameter `exclusive` defaults to `True` when using SQLite.
 
+        @param   db         Database instance
         @param   commit     whether transaction commits automatically at exiting with-block
         @param   exclusive  whether entering a with-block is exclusive
                             over other Transaction instances on this Database
