@@ -8,7 +8,7 @@ Released under the MIT License.
 
 @author      Erki Suurjaak
 @created     08.05.2020
-@modified    21.11.2022
+@modified    24.11.2022
 ------------------------------------------------------------------------------
 """
 import collections
@@ -71,15 +71,13 @@ class Queryable(api.Queryable):
         return next(iter(row.values())) if row and isinstance(row, dict) else None
 
 
-    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(),
-                limit=(), values=()):
+    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(), limit=(), values=()):
         """Returns (SQL statement string, parameter dict)."""
 
-        SCHEMA = self._load_schema()
-
         def cast(col, val):
             """Returns column value cast to correct type for use in psycopg."""
-            field = table in SCHEMA and SCHEMA[table]["fields"].get(col)
+            self._load_schema()
+            field = table in self._structure and self._structure[table]["fields"].get(col)
             if field and "array" == field["type"]:
                 return list(listify(val)) # Values for array fields must be lists
             elif field and val is not None:
@@ -109,13 +107,14 @@ class Queryable(api.Queryable):
                 col = "%s%s = ANY('{}')" % ("" if "IN" == op else "NOT ", col)
                 op = "EXPR"
             return col, op, val, key
-        def argcount(x): return len(x) if isinstance(x, (list, set, tuple)) else 1
-        def listify(x) : return x if isinstance(x, (list, tuple)) else [x]
+        def argcount(x)  : return len(x) if isinstance(x, (list, set, tuple)) else 1
+        def listify(x)   : return x if isinstance(x, (list, tuple)) else [x]
+        def strlistify(x): return [text_type(y) for y in listify(x) if y not in ("", None)]
 
         action = action.upper()
-        cols   =    cols if isinstance(cols,  string_types) else ", ".join(cols)
+        cols   =    cols if isinstance(cols,  string_types) else ", ".join(strlistify(cols)) or "*"
         where  = [where] if isinstance(where, string_types) else where
-        group  =   group if isinstance(group, string_types) else ", ".join(map(str, listify(group)))
+        group  =   group if isinstance(group, string_types) else ", ".join(strlistify(group))
         order  = [order] if isinstance(order, string_types) else order
         order  = [order] if isinstance(order, (list, tuple)) \
                  and len(order) == 2 and isinstance(order[1], bool) else order
@@ -133,8 +132,9 @@ class Queryable(api.Queryable):
             args.update((n, cast(k, v)) for n, (k, v) in zip(keys, values))
             cols, vals = ", ".join(k for k, _ in values), ", ".join("%%(%s)s" % n for n in keys)
             sql += " (%s) VALUES (%s)" % (cols, vals)
-            if SCHEMA and table in SCHEMA and SCHEMA[table].get("key"):
-                sql += " RETURNING %s AS id" % (SCHEMA[table]["key"])
+            self._load_schema()
+            if table in self._structure and self._structure[table].get("key"):
+                sql += " RETURNING %s AS id" % (self._structure[table]["key"])
         if "UPDATE" == action:
             sql += " SET "
             for i, (col, val) in enumerate(values):
@@ -169,9 +169,11 @@ class Queryable(api.Queryable):
             sql += " GROUP BY " + group
         if order:
             sql += " ORDER BY "
-            for i, col in enumerate(order):
-                name = col if isinstance(col, string_types) else col[0]
-                sort = col[1] if name != col and len(col) > 1 else ""
+            for i, col in enumerate(listify(order)):
+                name = col if isinstance(col, string_types) else \
+                       text_type(col[0] if isinstance(col, (list, tuple)) else col)
+                sort = col[1] if name != col and isinstance(col, (list, tuple)) and len(col) > 1 \
+                       else ""
                 if not isinstance(sort, string_types): sort = "DESC" if sort else ""
                 sql += (", " if i else "") + name + (" " if sort else "") + sort
         if limit:
@@ -190,6 +192,7 @@ class Queryable(api.Queryable):
         Returns identifier in quotes and proper-escaped for queries,
         if value needs quoting (has non-alphanumerics, starts with number, or is reserved).
 
+        @param   value  the value to quote, returned as-is if not string
         @param   force  whether to quote value even if not required
         """
         return quote(value, force)
@@ -206,11 +209,10 @@ class Queryable(api.Queryable):
 
 
     def _load_schema(self, force=False):
-        """Returns database table structure, queried from database if uninitialized or forced."""
+        """Populates table structure from database if uninitialized or forced."""
         if self._structure is None or force:
             self._structure = {}  # Avoid recursion on first query
             self._structure.update(query_schema(self, keys=True))
-        return self._structure
 
 
 
@@ -279,6 +281,7 @@ class Database(api.Database, Queryable):
         """
         Executes SQL statement, returns psycopg cursor.
 
+        @param   sql   SQL statement to execute, with psycopg-specific parameter bindings, if any
         @param   args  dictionary for %(name)s placeholders,
                        or a sequence for positional %s placeholders, or None
         """
@@ -292,9 +295,11 @@ class Database(api.Database, Queryable):
         Executes the SQL as script of any number of statements.
 
         Reloads internal schema structure from database.
+
+        @param   sql   script with one or more SQL statements
         """
         cursor = self.execute(sql)
-        self._load_schema(force=True)
+        self._structure = None  # Clear database schema to force reload on next query
         return cursor
 
 
@@ -303,7 +308,7 @@ class Database(api.Database, Queryable):
         if self._cursor: return
         self.init_pool(self, **self._kwargs)
         self._apply_converters()
-        self._cursorctx = self.get_cursor(autocommit=True)
+        self._cursorctx = self.make_cursor(autocommit=True)
         self._cursor    = self._cursorctx.__enter__()
 
 
@@ -332,6 +337,12 @@ class Database(api.Database, Queryable):
         return not self._cursor
 
 
+    @property
+    def cursor(self):
+        """Database engine cursor object, or `None` if closed."""
+        return self._cursor
+
+
     def transaction(self, commit=True, exclusive=False, **kwargs):
         """
         Returns a transaction context manager.
@@ -349,7 +360,7 @@ class Database(api.Database, Queryable):
 
 
     @contextmanager
-    def get_cursor(self, commit=False, autocommit=False, schema=None, lazy=False):
+    def make_cursor(self, commit=False, autocommit=False, schema=None, lazy=False, itersize=None):
         """
         Context manager for psycopg connection cursor.
         Creates a new cursor on an unused connection and closes it when exiting
@@ -358,8 +369,9 @@ class Database(api.Database, Queryable):
         @param   commit      commit at the end on success
         @param   autocommit  connection autocommit mode
         @param   schema      name of Postgres schema to use, if not using default `"public"`
-        @param   lazy        if true, returns a named cursor that fetches rows iteratively;
-                             only supports making a single query
+        @param   lazy        if true, returns a named server-side cursor that fetches rows
+                             iteratively in batches; only supports making a single query
+        @param   itersize    batch size in rows for server-side cursor
         @return              psycopg2.extras.RealDictCursor
         """
         connection = self.POOLS[self].getconn()
@@ -371,8 +383,9 @@ class Database(api.Database, Queryable):
             # If using schema, schema tables are queried first, fallback to public.
             # Need two cursors if schema+lazy, as named cursor only does one query.
             if schema or not lazy: cursor = connection.cursor()
-            if schema: cursor.execute('SET search_path TO "%s",public' % schema)
+            if schema: cursor.execute('SET search_path TO %s,public' % quote(schema))
             if lazy: namedcursor = connection.cursor("name_%s" % id(connection))
+            if lazy and itersize is not None: namedcursor.itersize = itersize
 
             try:
                 yield namedcursor or cursor
@@ -409,7 +422,7 @@ class Database(api.Database, Queryable):
         if not self.CONVERTERS: return
 
         regs, self.CONVERTERS = dict(self.CONVERTERS), {}
-        with self.get_cursor() as cursor:
+        with self.make_cursor() as cursor:
             for typename, transformer in regs.items():
                 cursor.execute("SELECT NULL::%s" % typename)
                 oid = cursor.description[0][1]  # description is [(name, type_code, ..)]
@@ -428,28 +441,33 @@ class Transaction(api.Transaction, Queryable):
     """
     Transaction context manager, provides convenience methods for queries.
 
-    Supports lazy cursors; those can only be used for making a single query.
+    Supports server-side cursors; those can only be used for making a single query.
 
     Must be closed explicitly if not used as context manager in a with-block.
     Block can be exited early by raising Rollback.
     """
 
-    def __init__(self, db, commit=True, exclusive=False, schema=None, lazy=False, **__):
+    def __init__(self, db, commit=True, exclusive=False,
+                 schema=None, lazy=False, itersize=None, **__):
         """
         Creates a transaction context manager.
 
         Context is breakable by raising Rollback.
 
+        @param   db         Database instance
         @param   commit     whether transaction commits automatically at exiting with-block
         @param   exclusive  whether entering a with-block is exclusive over other
                             Transaction instances Database
         @param   schema     search_path to use in this transaction
-        @param   lazy       if true, fetches results from server iteratively
-                            instead of all at once, supports single query only
+        @param   lazy       if true, uses a server-side cursor to fetch results from server
+                            iteratively in batches instead of all at once,
+                            supports one single query only
+        @param   itersize   batch size for server-side cursor (defaults to 2000 rows)
         """
         self._db         = db
+        self._lazy       = lazy
         self._cursor     = None
-        self._cursorctx  = db.get_cursor(commit=commit, schema=schema, lazy=lazy)
+        self._cursorctx  = db.make_cursor(commit, schema=schema, lazy=lazy, itersize=itersize)
         self._exclusive  = exclusive
         self._exitcommit = commit
         self._enterstack = 0     # Number of levels the transaction context is nested at
@@ -508,6 +526,7 @@ class Transaction(api.Transaction, Queryable):
         """
         Executes SQL statement, returns psycopg cursor.
 
+        @param   sql   SQL statement to execute, with psycopg-specific parameter bindings, if any
         @param   args  dictionary for %(name)s placeholders,
                        or a sequence for positional %s placeholders, or None
         """
@@ -521,9 +540,11 @@ class Transaction(api.Transaction, Queryable):
         Executes the SQL as script of any number of statements.
 
         Reloads internal schema structure from database.
+
+        @param   sql   script with one or more SQL statements
         """
         cursor = self.execute(sql)
-        self._load_schema(force=True)
+        self._structure = None  # Clear database schema to force reload on next query
         return cursor
 
     def commit(self):
@@ -539,11 +560,27 @@ class Transaction(api.Transaction, Queryable):
         """Whether transaction is currently not open."""
         return not self._cursorctx
 
+    @property
+    def cursor(self):
+        """Database engine cursor object, or `None` if closed."""
+        if not self._cursorctx: return None
+        if not self._cursor: self._cursor = self._cursorctx.__enter__()
+        return self._cursor
+
     @property
     def database(self):
         """Returns transaction Database instance."""
         return self._db
 
+    def _load_schema(self, force=False):
+        """
+        Populates database table structure from database if uninitialized or forced.
+
+        Uses parent Database for lookup if lazy cursor.
+        """
+        if self._lazy: return self._db._load_schema(force=force)
+        return super(Transaction, self)._load_schema(force=force)
+
 
 def autodetect(opts):
     """
@@ -572,7 +609,7 @@ def make_db_url(opts):
         if creds and i > 1: result, creds = result + "@", False  # Either user or password set
         if opts.get(k) is not None:
             result, creds = result + prefix + "%%(%s)s" % k, (i < 2)
-    result %= {k : urllib_parse.quote(str(opts[k])) for k in opts}
+    result %= {k : urllib_parse.quote(text_type(opts[k])) for k in opts}
     if any(k not in BASICS for k in opts):
         result += "/" if opts.get("dbname") is None else ""
         result += "?" + urllib_parse.urlencode({k: opts[k] for k in opts if k not in BASICS})
@@ -669,6 +706,7 @@ def quote(value, force=False):
     Returns identifier in quotes and proper-escaped for queries,
     if value needs quoting (has non-alphanumerics, starts with number, or is reserved).
 
+    @param   value  the value to quote, returned as-is if not string
     @param   force  whether to quote value even if not required
     """
     if not isinstance(value, string_types):
