@@ -8,7 +8,7 @@ Released under the MIT License.
 
 @author      Erki Suurjaak
 @created     05.03.2014
-@modified    21.11.2022
+@modified    25.11.2022
 ------------------------------------------------------------------------------
 """
 import collections
@@ -19,7 +19,7 @@ import sqlite3
 import sys
 import threading
 
-from six import binary_type, integer_types, string_types
+from six import binary_type, integer_types, string_types, text_type
 
 from .. import api
 
@@ -61,8 +61,7 @@ class Queryable(api.Queryable):
         return self.execute(sql, args).lastrowid
 
 
-    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(),
-                limit=(), values=()):
+    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(), limit=(), values=()):
         """Returns (SQL statement string, parameter dict)."""
 
         def cast(col, val):
@@ -87,13 +86,14 @@ class Queryable(api.Queryable):
                     # ("col", ("SQL with ? placeholders", val))
                     col, val, op = "%s = %s" % (col, val[0]), listify(val[1]), "EXPR"
             return col, op, val, key
-        def argcount(x): return len(x) if isinstance(x, (list, set, tuple)) else 1
-        def listify(x) : return x if isinstance(x, (list, tuple)) else [x]
+        def argcount(x)  : return len(x) if isinstance(x, (list, set, tuple)) else 1
+        def listify(x)   : return x if isinstance(x, (list, tuple)) else [x]
+        def strlistify(x): return [text_type(y) for y in listify(x) if y not in ("", None)]
 
         action = action.upper()
-        cols   =    cols if isinstance(cols,  string_types) else ", ".join(cols)
+        cols   =    cols if isinstance(cols,  string_types) else ", ".join(strlistify(cols)) or "*"
         where  = [where] if isinstance(where, string_types) else where
-        group  =   group if isinstance(group, string_types) else ", ".join(map(str, listify(group)))
+        group  =   group if isinstance(group, string_types) else ", ".join(strlistify(group))
         order  = [order] if isinstance(order, string_types) else order
         order  = [order] if isinstance(order, (list, tuple)) \
                  and len(order) == 2 and isinstance(order[1], bool) else order
@@ -150,9 +150,11 @@ class Queryable(api.Queryable):
             sql += " GROUP BY " + group
         if order:
             sql += " ORDER BY "
-            for i, col in enumerate(order):
-                name = col if isinstance(col, string_types) else col[0]
-                sort = col[1] if name != col and len(col) > 1 else ""
+            for i, col in enumerate(listify(order)):
+                name = col if isinstance(col, string_types) else \
+                       text_type(col[0] if isinstance(col, (list, tuple)) else col)
+                sort = col[1] if name != col and isinstance(col, (list, tuple)) and len(col) > 1 \
+                       else ""
                 if not isinstance(sort, string_types): sort = "DESC" if sort else ""
                 sql += (", " if i else "") + name + (" " if sort else "") + sort
         if limit:
@@ -173,6 +175,7 @@ class Queryable(api.Queryable):
         Returns identifier in quotes and proper-escaped for queries,
         if value needs quoting (has non-alphanumerics, starts with number, or is reserved).
 
+        @param   value  the value to quote, returned as-is if not string
         @param   force  whether to quote value even if not required
         """
         return quote(value, force)
@@ -194,6 +197,7 @@ class Database(api.Database, Queryable):
         """
         Creates a new Database instance for SQLite.
 
+        @param   opts    file path or `":memory:"`
         @param   kwargs  supported arguments are passed to sqlite3.connect() in open(),
                          like `detect_types=sqlite3.PARSE_COLNAMES`
         """
@@ -219,12 +223,22 @@ class Database(api.Database, Queryable):
 
 
     def execute(self, sql, args=()):
-        """Executes the SQL statement and returns sqlite3.Cursor."""
+        """
+        Executes the SQL statement and returns sqlite3.Cursor.
+
+        @param   sql   SQL statement to execute, with SQLite-specific parameter bindings, if any
+        @param   args  dictionary for %(name)s placeholders,
+                       or a sequence for positional %s placeholders
+        """
         return self.connection.execute(sql, args)
 
 
     def executescript(self, sql):
-        """Executes the SQL as script of any number of statements."""
+        """
+        Executes the SQL as script of any number of statements.
+
+        @param   sql   script with one or more SQL statements
+        """
         self.connection.executescript(sql)
 
 
@@ -264,6 +278,12 @@ class Database(api.Database, Queryable):
         return not self.connection
 
 
+    @property
+    def cursor(self):
+        """Database engine cursor object, or `None` if closed."""
+        return self.connection.cursor() if self.connection else None
+
+
     def transaction(self, commit=True, exclusive=True, **kwargs):
         """
         Returns a transaction context manager.
@@ -299,6 +319,7 @@ class Transaction(api.Transaction, Queryable):
         """
         Creates a new transaction.
 
+        @param   db         Database instance
         @param   commit     whether transaction commits automatically at exiting with-block
         @param   exclusive  whether entering a with-block is exclusive over other
                             Transaction instances on this Database
@@ -306,7 +327,7 @@ class Transaction(api.Transaction, Queryable):
         self._db         = db
         self._exitcommit = commit
         self._enterstack = 0  # Number of levels the transaction context is nested at
-        self._isolevel0  = None
+        self._isolevel0  = db.connection.isolation_level
         self._exclusive  = True if exclusive is None else exclusive
         self._closed     = False
         self._cursor     = None
@@ -316,13 +337,10 @@ class Transaction(api.Transaction, Queryable):
         if self._closed: raise RuntimeError("Transaction already closed")
 
         if self._exclusive: Database.MUTEX[self._db].acquire()
-        if not self._cursor:
-            self._isolevel0 = self._db.connection.isolation_level
-            self._db.connection.isolation_level = "DEFERRED"
-            try: self._cursor = self._db.execute("SAVEPOINT tx%s" % id(self))
-            except Exception:
-                if self._exclusive: Database.MUTEX[self._db].release()
-                raise
+        try: not self._cursor and self._make_cursor()
+        except Exception:
+            if self._exclusive: Database.MUTEX[self._db].release()
+            raise
         self._enterstack += 1
         return self
 
@@ -351,19 +369,23 @@ class Transaction(api.Transaction, Queryable):
         if self._closed:
             self._db._notify(self)
             return
-        if commit is False: self.rollback()
+        if commit is False or not commit and not self._exitcommit: self.rollback()
         elif commit or self._exitcommit: self.commit()
+        self._cursor = None
         self._closed = True
         self._db.connection.isolation_level = self._isolevel0
         self._db._notify(self)
 
     def execute(self, sql, args=()):
-        """Executes the SQL statement and returns sqlite3.Cursor."""
+        """
+        Executes the SQL statement and returns sqlite3.Cursor.
+
+        @param   sql   SQL statement to execute, with SQLite-specific parameter bindings, if any
+        @param   args  dictionary for %(name)s placeholders,
+                       or a sequence for positional %s placeholders
+        """
         if self._closed: raise RuntimeError("Transaction already closed")
-        if not self._cursor:
-            self._isolevel0 = self._db.connection.isolation_level
-            self._db.connection.isolation_level = "DEFERRED"
-            self._cursor = self._db.execute("SAVEPOINT tx%s" % id(self))
+        if not self._cursor: self._make_cursor()
         return self._cursor.execute(sql, args)
 
     def executescript(self, sql):
@@ -371,6 +393,8 @@ class Transaction(api.Transaction, Queryable):
         Executes the SQL as script of any number of statements, outside of transaction.
 
         Any pending transaction will be committed first.
+
+        @param   sql   script with one or more SQL statements
         """
         if self._closed: raise RuntimeError("Transaction already closed")
         with Database.MUTEX[self._db]:
@@ -394,11 +418,27 @@ class Transaction(api.Transaction, Queryable):
         """Whether transaction is currently not open."""
         return self._closed
 
+    @property
+    def cursor(self):
+        """Database engine cursor object, or `None` if closed."""
+        if self._closed: return None
+        if not self._cursor: self._make_cursor()
+        return self._cursor
+
+
     @property
     def database(self):
         """Returns transaction Database instance."""
         return self._db
 
+    def _make_cursor(self):
+        """Opens the transaction cursor."""
+        self._db.connection.isolation_level = "DEFERRED"
+        try: self._cursor = self._db.execute("SAVEPOINT tx%s" % id(self))
+        except Exception:
+            self._db.connection.isolation_level = self._isolevel0
+            raise
+
     def _reset(self, commit=False):
         """Commits or rolls back ongoing transaction, if any, closes cursor, if any."""
         if getattr(self._db.connection, "in_transaction", True):  # Py 3.2+
@@ -428,6 +468,7 @@ def quote(value, force=False):
     Returns identifier in quotes and proper-escaped for queries,
     if value needs quoting (has non-alphanumerics, starts with number, or is reserved).
 
+    @param   value  the value to quote, returned as-is if not string
     @param   force  whether to quote value even if not required
     """
     if not isinstance(value, string_types):
@@ -453,27 +494,3 @@ __all__ = [
     "RESERVED_KEYWORDS", "Database", "Transaction",
     "autodetect", "quote", "register_adapter", "register_converter",
 ]
-
-
-if "__main__" == __name__:
-    def test():
-        sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
-        import dblite as db
-        db.init(":memory:")
-        db.execute("CREATE TABLE test (id INTEGER PRIMARY KEY, val TEXT)")
-
-        print("Inserted ID %s." % db.insert("test", val=None))
-        for i in range(5): print("Inserted ID %s." % db.insert("test", {"val": i}))
-        print("Fetch ID 1: %s." % db.fetchone("test", id=1))
-        print("Fetch all up to 3, order by val: %s." % db.fetchall("test", order="val", limit=3))
-        print("Updated %s row where val is NULL." % db.update("test", {"val": "new"}, val=None))
-        print("Select where val IN [0, 1, 2]: %s." % db.fetchall("test", val=("IN", range(3))))
-        print("Delete %s row where val=0." % db.delete("test", val=0))
-        with db.transaction():
-            print("Delete %s row where val=1, and roll back." % db.delete("test", val=1))
-            raise db.Rollback
-        print("Fetch all, order by val: %s." % db.fetchall("test", order="val"))
-        print("Select with expression: %s." % db.fetchall("test", EXPR=("id IN (SELECT id FROM test WHERE id in (?, ?, ?))", [1, 3, 5])))
-        db.execute("DROP TABLE test")
-        db.close()
-    test()
