@@ -3,7 +3,7 @@
 """
 Test database general API in available engines.
 
-Running Postgres test needs sufficient variables in environment lik `PGUSER`.
+Running Postgres test needs sufficient variables in environment like `PGUSER`.
 
 ------------------------------------------------------------------------------
 This file is part of dblite - simple query interface for SQL databases.
@@ -11,7 +11,7 @@ Released under the MIT License.
 
 @author      Erki Suurjaak
 @created     20.11.2022
-@modified    22.11.2022
+@modified    24.11.2022
 ------------------------------------------------------------------------------
 """
 import collections
@@ -19,6 +19,8 @@ import copy
 import logging
 import os
 import tempfile
+import threading
+import time
 import unittest
 
 import dblite
@@ -91,7 +93,7 @@ class TestAPI(unittest.TestCase):
         """Tests dblite API."""
         logger.info("Verifying dblite API.")
         for engine, (opts, kwargs) in self._connections.items():
-            self.verify_module_api(opts, kwargs, engine)
+            self.verify_general_api(opts, kwargs, engine)
             self.verify_query_api(dblite, engine)
             dblite.close()
             with dblite.init() as db:
@@ -100,11 +102,12 @@ class TestAPI(unittest.TestCase):
                 self.verify_query_api(tx, engine)
             self.verify_query_args(dblite, engine)
             self.verify_transactions(engine)
+            self.verify_exclusive_transactions(engine)
             dblite.api.Engines.DATABASES.clear()  # Clear cache of default databases
 
 
-    def verify_module_api(self, opts, kwargs, engine):
-        """Verifies general module-level functions."""
+    def verify_general_api(self, opts, kwargs, engine):
+        """Verifies general module-level and Database-level functions."""
         logger.info("Verifying dblite module-level functions for %s.", engine)
 
         logger.debug("Verifying dblite.init().")
@@ -119,10 +122,15 @@ class TestAPI(unittest.TestCase):
                               "Unexpected value from dblite.transaction().")
         tx.close()
 
+        logger.info("Verifying Database property decorators for %s.", engine)
+        self.assertFalse(db.closed, "Unexpected value from %s.closed." % label(db))
+        self.assertIsNotNone(db.cursor, "Unexpected value from %s.cursor." % label(db))
+
         logger.debug("Verifying dblite.close().")
-        self.assertFalse(db.closed, "Unexpected value from Database.closed.")
+        self.assertFalse(db.closed, "Unexpected value from %s.closed." % label(db))
         dblite.close()
-        self.assertTrue(db.closed, "Unexpected value from Database.closed.")
+        self.assertTrue(db.closed, "Unexpected value from %s.closed." % label(db))
+        self.assertIsNone(db.cursor, "Unexpected value from %s.cursor." % label(db))
         with self.assertRaises(Exception,
                                msg="Unexpected success for fetch after closing database."):
             db.fetchone(next(iter(self.TABLES)))
@@ -198,6 +206,10 @@ class TestAPI(unittest.TestCase):
     def verify_query_args(self, obj, engine):
         """Verifies various ways of providing query parameters."""
         logger.info("Verifying %s query parameters for %s.", label(obj), engine)
+        class Column(object):
+            """Simple stringable class."""
+            def __init__(self, name): self._name = name
+            def __str__(self): return self._name
 
         for table, cols in self.TABLES.items():
             obj.executescript("DROP TABLE IF EXISTS %s" % table)
@@ -216,10 +228,12 @@ class TestAPI(unittest.TestCase):
         logger.info("Verifying SELECT columns for %r.", engine)
         for table, cols in self.TABLES.items():
             for col in (", ".join(sorted(c["name"] for c in cols)),
+                        [Column(c["name"]) for c in cols], None,
                         [c["name"] for c in cols], [c["name"] for c in cols][::2]):
                 row = obj.fetchone(table, col)
                 received = set(row) if isinstance(col, list) else ", ".join(sorted(row))
-                expected = set(col) if isinstance(col, list) else col
+                expected = set(map(str, col)) if isinstance(col, list) else \
+                           ", ".join(sorted(c["name"] for c in cols)) if col is None else col
                 self.assertEqual(received, expected,
                                  "Unexpected value from %s.select(cols)." % label(obj))
 
@@ -259,14 +273,14 @@ class TestAPI(unittest.TestCase):
         logger.info("Verifying ORDER BY arguments for %r.", engine)
         for table in DATAS:
             ORDERS = [  # [(argument value, [(col, direction), ])]
-                ("id",                           [("id",  False), ]),
-                ("id ASC",                       [("id",  False), ]),
-                ("id DESC",                      [("id",  True),  ]),
-                (["id", True],                   [("id",  True),  ]),
-                ("val, id DESC",                 [("val", False), ("id",  True)]),
-                (["val", "id DESC"],             [("val", False), ("id",  True)]),
-                (["val", ("id", "DESC")],        [("val", False), ("id",  True)]),
-                (["val DESC", ("id", True)],     [("val", True),  ("id",  True)]),
+                (Column("id"),                    [("id",  False), ]),
+                ("id ASC",                        [("id",  False), ]),
+                ("id DESC",                       [("id",  True),  ]),
+                ([Column("id"), True],            [("id",  True),  ]),
+                ("val, id DESC",                  [("val", False), ("id",  True)]),
+                ([Column("val"), "id DESC"],      [("val", False), ("id",  True)]),
+                (["val", (Column("id"), "DESC")], [("val", False), ("id",  True)]),
+                (["val DESC", ("id", True)],      [("val", True),  ("id",  True)]),
             ]
             for order, sorts in ORDERS:
                 reverse = "val" == sorts[0][0] and sorts[0][1]
@@ -280,7 +294,7 @@ class TestAPI(unittest.TestCase):
         logger.info("Verifying GROUP BY arguments for %r.", engine)
         for table in DATAS:
             expected_ids = all_ids = [x["id"] for x in DATAS[table]]
-            for group in ("id", "id, val", ["id"], ["id", "val"]):
+            for group in ("id", Column("id"), "id, val", ["id"], [Column("id"), "val"], 1):
                 rows = obj.fetchall(table, group=group)
                 self.assertEqual(set(x["id"] for x in rows), set(expected_ids),
                                  "Unexpected value from %s.select(group=%r)." % (label(obj), group))
@@ -378,6 +392,15 @@ class TestAPI(unittest.TestCase):
         for table in self.DATAS:
             dblite.execute("DROP TABLE %s" % table)
 
+        logger.info("Verifying Transaction property decorators for %s.", engine)
+        with dblite.transaction() as tx:
+            self.assertFalse(tx.closed, "Unexpected value from %s.closed." % label(tx))
+            self.assertIsNotNone(tx.cursor, "Unexpected value from %s.cursor." % label(tx))
+            self.assertIsInstance(tx.database, dblite.Database,
+                                  "Unexpected value from %s.database." % label(tx))
+        self.assertTrue(tx.closed, "Unexpected value from %s.closed." % label(tx))
+        self.assertIsNone(tx.cursor, "Unexpected value from %s.cursor." % label(tx))
+
         logger.info("Verifying Transaction.quote() for %s.", engine)
         with dblite.transaction() as tx:
             for value, same in [("WHERE", False), ("one two", False), ("abcd", True)]:
@@ -391,6 +414,25 @@ class TestAPI(unittest.TestCase):
                                         (label(tx), value, result))
 
 
+    def verify_exclusive_transactions(self, engine):
+        """Verifies exclusive transactions being exclusive."""
+        logger.info("Verifying exclusive transactions for %s.", engine)
+
+        DELAY = 3
+        def waiter(semaphore):
+            """Opens transaction and sleeps for a bit."""
+            with dblite.transaction(exclusive=True):
+                semaphore.set()
+                time.sleep(DELAY)
+
+        semaphore = threading.Event()
+        threading.Thread(target=waiter, args=(semaphore, )).start()
+        semaphore.wait()
+        t1 = time.time()
+        with dblite.transaction(exclusive=True):
+            t2 = time.time()
+        self.assertLessEqual(t1, t2 - DELAY, "Exclusive transaction did not exclude.")
+
 
 def label(obj):
     """Returns readable name for logging, for `dblite` module or class instances."""
@@ -399,6 +441,7 @@ def label(obj):
     return obj.__name__
 
 
+
 if "__main__" == __name__:
     logging.basicConfig(
         level=logging.INFO,
